import json
from functools import update_wrapper

from admin_ordering.admin import OrderableAdmin
from django import forms
from django.urls import re_path
from django.contrib import admin
from django.core.exceptions import PermissionDenied
from django.urls import reverse
from django.http import Http404
from django.shortcuts import redirect
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from formulaic import models as formulaic_models


def archive_forms(modeladmin, request, queryset):
    """
    Action: archive all selected forms
    """
    queryset.update(archived=True)


archive_forms.short_description = "Archive selected forms"


def form_status(form):
    """
    List Column: display static, currently Active/Archived
    """
    if form.archived:
        return mark_safe("Archived")
    else:
        return mark_safe("<strong>Active</strong>")


form_status.short_description = "Status"
form_status.allow_tags = True


def form_submissions(form):
    template = """
      <div style="line-height: 20px;">
        <a
          href="/formulaic/download/submissions/?form={pk}"
        >
            Download Submissions CSV
        </a>
      </div>
    """
    return mark_safe(template.format(
        pk=form.pk
    ))


form_submissions.short_description = u'Submissions'
form_submissions.allow_tags = True


def form_actions(form):
    """
    List Column: archive button for each form
    """
    if form.archived:
        url = reverse("admin:formulaic_form_unarchive", args=(form.pk,))
        return mark_safe('<a href="{}">Un-archive</a>'.format(url))
    else:
        url = reverse("admin:formulaic_form_archive", args=(form.pk,))
        return mark_safe('<a href="{}">Archive</a>'.format(url))


form_actions.short_description = ""
form_actions.allow_tags = True


class FormAdmin(admin.ModelAdmin):
    prepopulated_fields = {
        "slug": ("name",),
    }

    list_display = (
        "name",
        form_submissions,
        form_status,
        form_actions,
    )

    list_filter = (
        "archived",
    )

    search_fields = ('name',)

    actions = [archive_forms]

    @property
    def media(self):
        super_media = super(FormAdmin, self).media
        form_media = forms.Media(
            css={
                'all': (
                    '//ajax.googleapis.com/ajax/libs/jqueryui/1.11.4/themes/smoothness/jquery-ui.css',
                    'admin/formulaic/css/form.css',
                )
            },
            js=(
                '//ajax.googleapis.com/ajax/libs/jqueryui/1.11.4/jquery-ui.min.js',
                'admin/formulaic/js/form.js',
            )
        )

        return super_media + form_media

    def get_urls(self):
        def wrap(view):
            def wrapper(*args, **kwargs):
                return self.admin_site.admin_view(view)(*args, **kwargs)

            wrapper.model_admin = self
            return update_wrapper(wrapper, view)

        url_patterns = super().get_urls()

        return [
            re_path(r'^([0-9]+)/archive/$', wrap(self.archive_view),
                    name="formulaic_form_archive"),
            re_path(r'^([0-9]+)/unarchive/$', wrap(self.unarchive_view),
                    name="formulaic_form_unarchive"),
            # pattern eats remaining URL path used by `ember-formulaic`
            re_path(r'^([0-9]+)/.+$', wrap(self.changeform_view)),
        ] + url_patterns





from django.forms import fields, widgets
from django.test import TestCase

from formulaic import models
from formulaic.forms import CustomForm


class CustomFormTests(TestCase):

    def setUp(self):
        # create form
        form, created = models.Form.objects.get_or_create(name="My Test Form", slug="my-test-form")
        self.form_id = form.id

        # add textfield
        models.TextField.objects.get_or_create(
            name="Text Field 1",
            slug="text-field-1",
            required=True,
            help_text=None,
            # model_class="",
            # content_type=None,
            position=0,
            form_id=self.form_id,
            enabled=0, # todo: not implemented?
            css_class=None,
            subtype=models.TextField.SUBTYPE_TEXT,
        )

    def test_generation(self):
        """
        Test CustomForm init, fields, etc
        """
        field_count = 1

        form = models.Form.objects.get(pk=self.form_id)
        formulaic_form = CustomForm(instance_id="primary_form", form=form, label_suffix="", request=None)

        self.assertTrue(
            len(formulaic_form.fields) == field_count,
            "Form's field count is incorrect: {} instead of {}".format(
                len(formulaic_form.fields),
                field_count
            )
        )

        for field in formulaic_form:
            # TODO: this needs to be modified for different field types
            self.assertIsInstance(
                field.field,
                fields.CharField,
                "CustomForm generated wrong Django field type: {} instead of {}".format(
                    field.field.__class__,
                    fields.CharField
                )
            )
            self.assertIsInstance(
                field.field.widget,
                widgets.TextInput,
                "CustomForm generated wrong widget type: {} instead of {}".format(
                    field.field.widget.__class__,
                    widgets.TextInput
                )
            )

    def test_submission(self):
        """
        Test CustomForm submission
        """
        field_count = 1

        post_data = {
            "text-field-1": "My Test Value"
        }

        form = models.Form.objects.get(pk=self.form_id)
        custom_form = CustomForm(
            post_data,
            instance_id="primary_form",
            form=form,
            label_suffix="",
            request=None
        )

        self.assertTrue(custom_form.is_valid(), "Form isn't valid.")

        if custom_form.is_valid():
            obj = form.create_submission(custom_form.cleaned_data, source=custom_form.instance_id)

            submission = models.Submission.objects.get(pk=obj.pk)

            self.assertEqual(submission.form_id, self.form_id)
            self.assertIsNotNone(submission.date_created)
            self.assertEqual(submission.source, "primary_form")
            self.assertEqual(len(submission.custom_data.keys()), field_count)
            self.assertEqual(submission.custom_data["text-field-1"], "My Test Value")

    def test_state_and_city_auto_populated_from_zip(self):
        # create form
        zip_form, created = models.Form.objects.get_or_create(name="My Zip Form", slug="my-zip-form")
        post_data = {
            "Zipcode": "21043",
            "City": "",
            "State": ""
        }

        # add textfield
        models.TextField.objects.get_or_create(
            name="Zipcode",
            slug="Zipcode",
            required=True,
            help_text=None,
            position=0,
            form_id=zip_form.id,
            enabled=0,
            css_class=None,
            subtype=models.TextField.SUBTYPE_TEXT,
        )

        models.HiddenField.objects.get_or_create(
            name="City",
            slug="City",
            required=True,
            help_text=None,
            position=0,
            form_id=zip_form.id,
            enabled=0,
            css_class=None,
            subtype=models.Field.TYPE_HIDDEN
        )

        models.HiddenField.objects.get_or_create(
            name="State",
            slug="State",
            required=True,
            help_text=None,
            position=0,
            form_id=zip_form.id,
            enabled=0,
            css_class=None,
            subtype=models.Field.TYPE_HIDDEN
        )

        custom_form = CustomForm(
            post_data,
            instance_id="primary_form",
            form=zip_form,
            label_suffix="",
            request=None
        )

        self.assertTrue(custom_form.is_valid(), "Zip form isn't valid.")

        if custom_form.is_valid():
            obj = zip_form.create_submission(custom_form.cleaned_data, source=custom_form.instance_id)

            submission = models.Submission.objects.get(pk=obj.pk)

            self.assertEqual(submission.custom_data["State"], "Maryland")
            self.assertEqual(submission.custom_data["City"], "Ellicott City")



import json

from ckeditor.fields import RichTextField
import django
from django.contrib.contenttypes.models import ContentType
from django.db import models, transaction
from django.db.models import Max, Prefetch
from django.forms import fields, widgets
from django.utils import timezone
from django.utils.functional import cached_property
import phonenumbers
from six import iteritems, python_2_unicode_compatible, u

from formulaic import fields as custom_fields
from formulaic.auto_populate import attempt_kv_auto_populate
from formulaic.signals import submission_complete
from formulaic.validators import validate_mixed_content, validate_phone_number
from formulaic.widgets import PhoneInput



class Form(models.Model):
    BASE_COLUMN_HEADERS = ['date', 'source', 'promo_source']
    name = models.CharField(max_length=500)
    slug = models.SlugField(max_length=200)
    success_message = models.TextField(null=True, blank=True)
    privacy_policy = models.ForeignKey(
        'PrivacyPolicy', on_delete=models.PROTECT, null=True, blank=True
    )

    archived = models.BooleanField(default=False)

    def create_submission(self, cleaned_data, source=None, metadata=None, promo_source=None):

        with transaction.atomic():
            # case submission
            submission = Submission()
            submission.form = self
            submission.date_created = timezone.now()
            submission.source = source
            submission.metadata = metadata or {}
            submission.promo_source = promo_source
            submission.save()

            # key value pairs
            key_value_pairs = []
            for key, value in iteritems(cleaned_data):
                if isinstance(value, dict):
                    # handle multi-value fields
                    # todo: DRY
                    for sub_key, sub_value in iteritems(value):
                        # if None, use regular key
                        if sub_key:
                            final_sub_key = "{}[{}]".format(key, sub_key)
                        else:
                            final_sub_key = key

                        key_value = SubmissionKeyValue()
                        key_value.submission = submission
                        key_value.key = final_sub_key
                        key_value.field = self.get_field_by_slug(key)
                        key_value.value = sub_value
                        key_value_pairs.append(key_value)
                else:
                    # handle standard fields
                    # todo: DRY
                    key_value = SubmissionKeyValue()
                    key_value.submission = submission
                    key_value.key = key
                    key_value.field = self.get_field_by_slug(key)
                    key_value.value = attempt_kv_auto_populate(key_value, value, cleaned_data)
                    key_value_pairs.append(key_value)

            SubmissionKeyValue.objects.bulk_create(key_value_pairs)

            submission_complete.send(sender=self.__class__, submission=submission, form=self)

        return submission

    @cached_property
    def column_headers(self):
        return (
            Form.BASE_COLUMN_HEADERS +
            list(self.field_set.all().order_by('position').values_list("slug", flat=True))
        )

    def get_field_by_slug(self, slug):
        if not hasattr(self, "_fields_dict"):
            all_fields = self.field_set.all()
            self._fields_dict = dict([(field.slug, field) for field in all_fields])

        return self._fields_dict.get(slug)

    @staticmethod
    def autocomplete_search_fields():
        return ("id__iexact", "name__icontains", "slug__icontains", )

    def __str__(self):
        if self.archived:
            return u("{} (archived)".format(self.name))
        else:
            return self.name

    class Meta:
        ordering = ('archived', 'name',)



class PrivacyPolicy(models.Model):
    """
    Provides an editable list of privacy policies which can be selected
    on any Formulaic Form.  Ideally, it will act to override the privacy
    policy on the page the form is rendered on.
    """

    name = models.CharField(max_length=250)
    text = RichTextField(
        config_name="very_basic",
        validators=[validate_mixed_content, ]
    )

    class Meta:
        verbose_name_plural = "Privacy policies"

    def __str__(self):
        return self.name



class OptionList(models.Model):
    """
    Collection of options for use in selects, checkbox lists,
    radio lists, etc...
    """

    name = models.CharField(max_length=250)

    @cached_property
    def cached_groups(self):
        return self.groups.all()

    @cached_property
    def has_groups(self):
        """
        Checks if OptionList has any OptionGroups.  Using len()
        instead of count() because its anticipated to be used
        in conjunction with cached_groups anyway.
        """
        return len(self.cached_groups) > 0

    def __str__(self):
        return self.name



class Option(models.Model):
    """
    An individual selectable option, represented as a member
    of an OptionList
    """

    name = models.CharField(max_length=250)
    value = models.CharField(max_length=250)
    position = models.PositiveIntegerField("Position", default=0)

    list = models.ForeignKey(OptionList, on_delete=models.CASCADE)

    def __str__(self):
        return self.name

    class Meta:
        ordering = ('position',)

    def save(self, *args, **kwargs):
        if not self.position:
            max = self.objects.aggregate(m=Max("position"))["m"]
            self.position = 10 + (max or 0)
        super(Option, self).save(*args, **kwargs)



class OptionGroup(models.Model):
    """
    A group of Options in an OptionList.  OptionGroups provide
    for flexibility to change which Options display based on
    the values of other fields via the Rules feature
    """

    name = models.CharField(max_length=250)
    position = models.PositiveIntegerField("Position", default=0)

    options = models.ManyToManyField(Option, related_name="groups")
    list = models.ForeignKey(
        OptionList, on_delete=models.CASCADE, related_name="groups"
    )

    @cached_property
    def cached_options(self):
        return self.options.all()

    def __str__(self):
        return "{}:{}".format(self.list.name, self.name)

    class Meta:
        ordering = ('position',)

    def save(self, *args, **kwargs):
        if not self.position:
            max = self.objects.aggregate(m=Max("position"))["m"]
            self.position = 10 + (max or 0)
        super(OptionGroup, self).save(*args, **kwargs)



class Field(models.Model):
    # TODO: look into changing these
    TYPE_SELECT = "select"
    TYPE_TEXT = "text"
    TYPE_CHECKBOX = "checkbox"
    TYPE_HIDDEN = "hidden"

    # displayed attributes
    name = models.CharField(max_length=500)  # TODO: remove?
    display_name = models.CharField(max_length=1000)
    data_name = models.CharField(max_length=200)
    slug = models.SlugField(max_length=200)
    required = models.BooleanField()
    help_text = models.CharField(max_length=500, blank=True, null=True)

    subtype = models.CharField(max_length=100)

    # bookkeeping attributes
    model_class = models.CharField(max_length=100)
    position = models.PositiveIntegerField("Position")
    form = models.ForeignKey(Form, on_delete=models.CASCADE)
    enabled = models.BooleanField(default=True)

    # styling attributes
    css_class = models.CharField(max_length=120, blank=True, null=True)

    # TODO: marked for removal
    content_type = models.ForeignKey(ContentType, on_delete=models.PROTECT)

    def subtype_is(self, subtype):
        return self.subtype == subtype

    def subtype_in(self, subtypes):
        return self.subtype in subtypes

    # TODO: delete this and change references to complete_field
    def get_specific_field(self):
        return self.content_type.get_object_for_this_type(id=self.id)

    @property
    def complete_field(self):
        model = self.content_type.model
        return getattr(self, model)

    class Meta:
        ordering = ['position']

        # TODO: based on requirements, this should probably be removed
        # might need to support duplicates across multiple pages
        unique_together = ("form", "slug")

    def __str__(self):
        return self.data_name

    def save(self, **kwargs):
        # TODO: temporary; testing Ember
        self.content_type = ContentType.objects.get_for_model(type(self))
        self.model_class = self.__class__.__name__.lower()

        super(Field, self).save(**kwargs)


class TextField(Field):
    SUBTYPE_TEXT = u"text"
    SUBTYPE_TEXTAREA = u"textarea"
    SUBTYPE_FULL_NAME = u"full_name"
    SUBTYPE_EMAIL = u"email"
    SUBTYPE_PHONE_NUMBER = u"phone_number"
    SUBTYPE_INTEGER = u"integer"

    SUBTYPES = {
        SUBTYPE_TEXT: {
            u"field_class": fields.CharField,
            u"widget_class": widgets.TextInput
        },
        SUBTYPE_TEXTAREA: {
            u"field_class": fields.CharField,
            u"widget_class": widgets.Textarea
        },
        SUBTYPE_EMAIL: {
            u"field_class": fields.EmailField,
            u"widget_class": widgets.TextInput
        },
        SUBTYPE_PHONE_NUMBER: {
            u"field_class": fields.CharField,
            u"widget_class": PhoneInput,
            u"validators": [validate_phone_number],
        },
        SUBTYPE_INTEGER: {
            u"field_class": fields.IntegerField,
            u"widget_class": widgets.TextInput
        },
        SUBTYPE_FULL_NAME: {
            u"field_class": custom_fields.FullNameField,
            u"widget_class": widgets.TextInput
        },
    }

    textarea_rows = models.PositiveIntegerField(blank=True, null=True)

    def get_implementation(self, widget_attrs=None):

        widget_attrs = widget_attrs or dict()

        subtype_options = TextField.SUBTYPES[self.subtype]

        widget_attrs[u"data-id"] = self.id

        if self.subtype == TextField.SUBTYPE_TEXTAREA:
            widget_attrs[u"rows"] = str(self.textarea_rows if self.textarea_rows else 4)

        widget_class = subtype_options[u"widget_class"]
        widget = widget_class(attrs=widget_attrs)

        field_class = subtype_options[u"field_class"]

        validators = subtype_options.get(u"validators", [])

        return field_class(
            label=self.display_name,
            required=self.required,
            widget=widget,
            validators=validators
        )

    def save(self, **kwargs):
        self.content_type = ContentType.objects.get_for_model(type(self))
        self.model_class = self.__class__.__name__.lower()

        super(TextField, self).save(**kwargs)


class HiddenField(Field):
    field_type = Field.TYPE_HIDDEN

    value = models.CharField(max_length=500, null=True, blank=True)

    def get_implementation(self, widget_attrs={}):
        widget_attrs[u"data-id"] = self.id
        widget = widgets.HiddenInput(attrs=widget_attrs)

        return fields.CharField(
            widget=widget,
            initial=self.value,
            required=False
        )


class BooleanField(Field):
    field_type = Field.TYPE_CHECKBOX

    default_checked = models.BooleanField()

    def get_implementation(self, widget_attrs={}):
        widget_attrs[u"data-id"] = self.id
        widget = widgets.CheckboxInput(attrs=widget_attrs)

        return fields.BooleanField(
            label=self.display_name,
            required=self.required,
            widget=widget,
            initial=self.default_checked
        )

    def save(self, **kwargs):
        self.content_type = ContentType.objects.get_for_model(type(self))
        self.model_class = self.__class__.__name__.lower()

        super(BooleanField, self).save(**kwargs)


class ChoiceField(Field):
    DEFAULT_TEXT_BACKUP = u"(Choose One)"

    SUBTYPE_SELECT = u"select"
    SUBTYPE_SELECTMULTIPLE = u"select_multiple"
    SUBTYPE_RADIOSELECT = u"radio_select"
    SUBTYPE_CHECKBOXSELECTMULTIPLE = u"checkbox_select_multiple"

    SUBTYPES = {
        SUBTYPE_SELECT: {
            u"field_class": fields.ChoiceField,
            u"widget_class": widgets.Select,
            u"multiple": False,
        },
        SUBTYPE_SELECTMULTIPLE: {
            u"field_class": fields.MultipleChoiceField,
            u"widget_class": widgets.SelectMultiple,
            u"multiple": True,
        },
        SUBTYPE_RADIOSELECT: {
            u"field_class": fields.ChoiceField,
            u"widget_class": widgets.RadioSelect,
            u"multiple": False,
        },
        SUBTYPE_CHECKBOXSELECTMULTIPLE: {
            u"field_class": fields.MultipleChoiceField,
            u"widget_class": widgets.CheckboxSelectMultiple,
            u"multiple": True,
        },
    }

    # multiselect = models.BooleanField()
    minimum_selections = models.PositiveIntegerField(blank=True, null=True)
    maximum_selections = models.PositiveIntegerField(blank=True, null=True)

    option_list = models.ForeignKey(OptionList, on_delete=models.PROTECT)
    option_group = models.ForeignKey(
        OptionGroup, on_delete=models.PROTECT, blank=True, null=True
    )
    default_options_string = models.CharField(max_length=200, blank=True, null=True)
    default_text = models.CharField(max_length=200, blank=True, null=True)

    field_type = Field.TYPE_SELECT
    field_class = fields.ChoiceField

    @property
    def needs_default_text(self):

        # Radio selects are a little special, in that they are multi selects,
        # but do not have multiple options. Having a default option does not
        # make sense.
        if self.subtype == self.SUBTYPE_RADIOSELECT:
            return False

        return not self.supports_multiple_values and not self.default_option

    @property
    def default_text_or_backup(self):
        return self.default_text or ChoiceField.DEFAULT_TEXT_BACKUP

    @property
    def supports_multiple_values(self):
        return ChoiceField.SUBTYPES[self.subtype][u"multiple"]

    @property
    def default_option(self):
        if self.default_options_string and not self.supports_multiple_values:
            return json.loads(self.default_options_string)[0]
        else:
            return None

    @default_option.setter
    def default_option(self, value):
        if not self.supports_multiple_values:
            if value:
                self.default_options_string = json.dumps([value])
            else:
                self.default_options_string = None

    @property
    def default_options(self):
        if self.default_options_string and self.supports_multiple_values:
            return json.loads(self.default_options_string)
        else:
            return []

    @default_options.setter
    def default_options(self, value):
        if self.supports_multiple_values:
            if value:
                self.default_options_string = json.dumps(value)
            else:
                self.default_options_string = None

    @property
    def options(self):
        if self.option_group:
            return self.option_group.options.all()
        else:
            return self.option_list.option_set.all()

    def get_implementation(self, widget_attrs={}):
        subtype_options = ChoiceField.SUBTYPES[self.subtype]

        widget_attrs[u"data-id"] = self.id
        widget_class = subtype_options[u"widget_class"]

        field_class = subtype_options[u"field_class"]

        default_options = []
        if self.needs_default_text:
            default_options.append(("", self.default_text_or_backup))
        for option in self.options:
            default_options.append((option.id, option.name))

        if self.supports_multiple_values and self.default_options:
            initial_value = self.default_options
        elif not self.supports_multiple_values and self.default_option:
            initial_value = self.default_option
        else:
            initial_value = None

        if not self.option_list.has_groups:
            # Standard implementation
            return field_class(
                label=self.display_name,
                required=self.required,
                choices=default_options,
                initial=initial_value,
                widget=widget_class(attrs=widget_attrs)
            )
        else:
            # GroupedChoiceField implementation
            groups = {
                None: default_options
            }

            for group in self.option_list.cached_groups:
                group_options = []

                if self.needs_default_text:
                    group_options.append((u"", self.default_text_or_backup))

                for option in group.cached_options:
                    group_options.append((option.id, option.name))

                groups[group.id] = group_options

            return custom_fields.GroupedChoiceField(
                label=self.display_name,
                required=self.required,
                groups=groups,
                initial=initial_value,
                field_class=field_class,
                widget_class=widget_class,
                widget_attrs=widget_attrs
            )

    def save(self, **kwargs):
        self.content_type = ContentType.objects.get_for_model(type(self))
        self.model_class = self.__class__.__name__.lower()

        super(ChoiceField, self).save(**kwargs)



class RuleResult(models.Model):
    ACTION_SHOW = 'show'
    ACTION_HIDE = 'hide'
    ACTION_REQUIRE = 'require'
    ACTION_OPTIONAL = 'optional'
    ACTION_CHANGE_OPTION_GROUP = 'change-option-group'

    ACTION_CHOICES = (
        (ACTION_SHOW, 'Show'),
        (ACTION_HIDE, 'Hide'),
        (ACTION_REQUIRE, 'Require (Override)'),
        (ACTION_OPTIONAL, 'Optional (Override)'),
        (ACTION_CHANGE_OPTION_GROUP, 'Change Option Group'),
    )

    action = models.CharField(max_length=50, choices=ACTION_CHOICES)
    field = models.ForeignKey(Field, on_delete=models.PROTECT)
    rule = models.ForeignKey(
        'Rule', on_delete=models.CASCADE, related_name='results', blank=True, null=True
    )
    option_group = models.ForeignKey(
        'OptionGroup', on_delete=models.PROTECT, blank=True, null=True
    )

    def __str__(self):
        return u("{}: '{}' field '{}' if rule '{}' is true".format(
            self.id,
            self.action,
            self.field_id,
            self.rule_id
        ))



class Rule(models.Model):
    OPERATOR_AND = 'and'
    OPERATOR_OR = 'or'
    OPERATOR_CHOICES = (
        (OPERATOR_AND, 'And'),
        (OPERATOR_OR, 'Or')
    )

    form = models.ForeignKey(Form, on_delete=models.CASCADE)
    # rule_set
    operator = models.CharField(max_length=3, choices=OPERATOR_CHOICES)
    position = models.IntegerField()

    def __str__(self):
        return u('{}: position "{}"'.format(self.id, self.position))

    class Meta:
        ordering = ('position',)



class RuleCondition(models.Model):
    OPERATOR_IS = 'is'
    OPERATOR_IS_NOT = 'is_not'
    OPERATOR_CONTAINS = 'contains'
    OPERATOR_DOES_NOT_CONTAIN = 'does_not_contain'
    OPERATOR_BEGINS_WITH = 'begins_with'
    OPERATOR_ENDS_WITH = 'ends_with'
    OPERATOR_GREATER_THAN = 'greater_than'
    OPERATOR_LESS_THAN = 'less_than'
    OPERATOR_ANY_SELECTED = 'any_selected'
    OPERATOR_ALL_SELECTED = 'all_selected'

    OPERATOR_CHOICES = (
        (OPERATOR_IS, 'is'),
        (OPERATOR_IS_NOT, 'is_not'),
        (OPERATOR_CONTAINS, 'contains'),
        (OPERATOR_DOES_NOT_CONTAIN, 'does_not_contain'),
        (OPERATOR_BEGINS_WITH, 'begins_with'),
        (OPERATOR_ENDS_WITH, 'ends_with'),
        (OPERATOR_GREATER_THAN, 'greater_than'),
        (OPERATOR_LESS_THAN, 'less_than'),
        (OPERATOR_ANY_SELECTED, 'any_selected'),
        (OPERATOR_ALL_SELECTED, 'all_selected')
    )

    position = models.PositiveIntegerField()
    rule = models.ForeignKey(
        Rule, on_delete=models.CASCADE, related_name="conditions", blank=True, null=True
    )

    field = models.ForeignKey(Field, on_delete=models.PROTECT)
    operator = models.CharField(max_length=30, choices=OPERATOR_CHOICES)

    value_string = models.TextField()

    @property
    def value(self):
        return json.loads(self.value_string)

    @value.setter
    def value(self, value):
        self.value_string = json.dumps(value)

    def __str__(self):
        return u('{}: field "{}" {} ______'.format(
            self.id,
            self.field_id,
            self.operator
        ))

    class Meta:
        ordering = ('position',)



class DisplayCondition(models.Model):
    IS = "is"
    IS_NOT = "is_not"

    VALUE_OPTIONS = (
        (IS, "is"),
        (IS_NOT, "is not")
    )

    values = models.TextField()
    value_option = models.CharField(max_length=15, choices=VALUE_OPTIONS)

    affected_field = models.ForeignKey(
        Field, on_delete=models.PROTECT, related_name="affecting_conditions"
    )
    watched_field = models.ForeignKey(
        Field, on_delete=models.PROTECT, related_name="watching_conditions"
    )

    def is_met_by_default(self):
        specific_watched_field = self.watched_field.get_specific_field()

        value_is = self.value_option == DisplayCondition.IS
        if specific_watched_field.field_type == Field.TYPE_SELECT:
            default_option_in_values = specific_watched_field.default_option in json.loads(self.values)
            return default_option_in_values if value_is else not default_option_in_values
        else:
            # TODO: setup default values which could match the DisplayCondition values
            return False

    def __str__(self):
        return "Display field #{} if field #{} {} {}".format(
            self.affected_field_id,
            self.watched_field_id,
            self.value_option,
            self.values
        )


class SubmissionQuerySet(models.QuerySet):

    _should_prefetch_custom_data = False
    _prefetch_custom_data_done = False

    def _clone(self, **kwargs):
        """Overloading _clone to allow for lazy prefetching of prefetch_custom_data

        This is required to make the prefetch_custom_data method commutative
        """

        # Compatibility
        if django.get_version() < "2.0.0":
            clone = super()._clone(**kwargs)
        else:
            clone = super()._clone()
        clone._should_prefetch_custom_data = self._should_prefetch_custom_data
        return clone

    def _fetch_all(self):
        """Overloading _fetch_all to allow for lazy prefetching of prefetch_custom_data"""

        r = super()._fetch_all()
        if self._should_prefetch_custom_data and not self._prefetch_custom_data_done:
            self._execute_prefetch_custom_data()
            self._prefetch_custom_data_done = True
        return r

    def _execute_prefetch_custom_data(self):
        """Performs the prefetch on the custom data in bulk
        """

        # First, determine what is the smallest set of options that we need to
        # create and then generate the options lookup dictionary.
        relevant_choices = (
            SubmissionKeyValue.objects
            .filter(submission__in=self, field__subtype__in=ChoiceField.SUBTYPES.keys())
            .values_list("value_charfield", flat=True)
            .order_by("value_charfield").distinct()
        )

        relevant_choices = [json.loads(choice) for choice in relevant_choices]
        # We need to "flatten" relevant choices, some entries could be lists.
        cleaned_relevant_choices = []
        for choice in relevant_choices:
            # Ignore blanks.
            if not choice:
                continue

            # Extend our cleaned list of relevant choices.
            # Make sure we only count integers.
            if isinstance(choice, list):
                cleaned_relevant_choices.extend([int(val) for val in choice if val.isdigit()])

            # These should already be integers. But better to make sure nothing
            # is slipping through.
            else:
                cleaned_relevant_choices.append(int(choice))

        # Remove duplicates
        cleaned_relevant_choices = set(cleaned_relevant_choices)

        # Create the options lookup dictionary.
        options = (
            Option.objects
            .filter(id__in=cleaned_relevant_choices)
            .values_list("id", "name")
        )
        options_lookup = dict(options)

        forms_by_id = dict()

        # We need to assign forms here to the python objects. Without this step,
        # we would create multiple python objects for the same form contained in
        # the database. By reassigning the form here we cache results onto a
        # single python object. It should be noted that in practice, there will
        # likely only be one form, but this supports submissions across
        # different forms.
        for submission in self._result_cache:
            if submission.form_id in forms_by_id:
                submission.form = forms_by_id[submission.form_id]
            else:
                forms_by_id[submission.form_id] = submission.form

            # Actually constructing the custom data and storing it.
            submission._get_custom_data_with_option_lookup(options_lookup)

    def prefetch_custom_data(self):
        """Returns a list with all of the custom data included."""

        self._should_prefetch_custom_data = True
        prefetch_values_qs = SubmissionKeyValue.objects.select_related("field")
        return self.select_related("form").prefetch_related(Prefetch("values", queryset=prefetch_values_qs))


class SubmissionManager(models.Manager):
    def get_queryset(self):
        return SubmissionQuerySet(self.model, using=self._db)

    def prefetch_custom_data(self):
        return self.get_queryset().prefetch_custom_data()


class Submission(models.Model):

    objects = SubmissionManager()

    form = models.ForeignKey(Form, on_delete=models.PROTECT)
    date_created = models.DateTimeField()

    source = models.CharField(
        max_length=200,
        null=True,
        blank=True,
        help_text="Name of the specific form placement the user filled out (e.g. sidebar-subscribe)"
    )

    promo_source = models.CharField(
        max_length=200,
        null=True,
        blank=True,
        help_text="Source passed through metadata variable promo_source"
    )

    @cached_property
    def custom_data(self):
        column_headers = self.form.column_headers
        data = {}

        for key_value in self.values.all():
            if key_value.key in column_headers:
                data[key_value.key] = key_value.output_value

        return data

    _prefetched_custom_data_cache = None

    @property
    def prefetched_custom_data(self):
        """Retrieves the custom data that was prefetched by the
        "prefetch_custom_data" queryset method.
        """

        if not self._prefetched_custom_data_cache:
            self._prefetched_custom_data_cache = self.custom_data
        return self._prefetched_custom_data_cache

    def _get_custom_data_with_option_lookup(self, options_lookup):
        """Similar output to "custom_data" property, however this allows for an
        options_lookup and importantly calls the _output_value_with_options_lookup
        method on the SubmissionKeyValue objects. Calling this with the lookup
        allows for much more efficient bulk exports.
         """
        if not self._prefetched_custom_data_cache:

            column_headers = self.form.column_headers
            data = {}

            for key_value in self.values.all():
                if key_value.key in column_headers:
                    data[key_value.key] = key_value._output_value_with_options_lookup(options_lookup)

            self._prefetched_custom_data_cache = data

        return self._prefetched_custom_data_cache

    metadata_serialized = models.TextField(
        help_text="Serialized JSON object storing arbitrary submission-related metadata"
    )

    @property
    def metadata(self):
        try:
            return json.loads(self.metadata_serialized)
        except ValueError:
            return {}

    @metadata.setter
    def metadata(self, value):
        self.metadata_serialized = json.dumps(value)



class SubmissionKeyValue(models.Model):
    submission = models.ForeignKey(
        Submission, on_delete=models.CASCADE, related_name="values"
    )
    key = models.CharField(max_length=200, db_index=True)
    value_charfield = models.CharField(max_length=500, null=True, blank=True)
    value_textfield = models.TextField(null=True, blank=True)
    field = models.ForeignKey(Field, on_delete=models.SET_NULL, null=True, blank=True)

    # TODO: cache output_value in database to speed up reports.  problem, option values can change...hmmm....

    @property
    def value(self):
        if self.value_charfield is not None:
            return json.loads(self.value_charfield)
        else:
            return self.value_textfield

    @value.setter
    def value(self, value):
        if self.field and self.field.subtype_is(TextField.SUBTYPE_TEXTAREA):
            self.value_textfield = value
            self.value_charfield = None
        else:
            self.value_textfield = None
            self.value_charfield = json.dumps(value)

    @property
    def output_value(self):
        value = self.value

        # Special Formatting for Phone Numbers
        if self.field and self.field.subtype_is(TextField.SUBTYPE_PHONE_NUMBER):
            try:
                parsed_number = phonenumbers.parse(value)
                return phonenumbers.format_number(
                    parsed_number, phonenumbers.PhoneNumberFormat.INTERNATIONAL
                )

            except phonenumbers.NumberParseException:
                # If phonenumbers doesn't know how to handle it, just return the
                # value as is.
                return value

        # replace ids with text values
        if value and self.field and self.field.subtype_in(ChoiceField.SUBTYPES.keys()):
            # convert to list for query
            if not isinstance(value, list):
                value = [value]

            selected_options = Option.objects.filter(id__in=value).values_list('name', flat=True)

            value = ",".join(selected_options)

        return value

    def _output_value_with_options_lookup(self, options_lookup):
        """Mirrors "output_value", but with an option_lookup dictionary. This
        dictionary avoids looking up the options in the database. This is useful for when
        exporting multiple submissions of the same form, as the Options being
        looked up are almost always the same.

        options_lookup is a dictionary with "values" as keys and "names" as values
        """
        value = self.value

        # replace ids with text values
        if value and self.field and self.field.subtype_in(ChoiceField.SUBTYPES.keys()):
            # convert to list for query
            if not isinstance(value, list):
                value = [value]

            selected_options = [options_lookup.get(json.loads(v)) for v in value]
            return ",".join(selected_options)

        return self.output_value

    def __str__(self):
        return "{}:{}".format(self.key, self.value)



from django import template
from django.conf import settings

register = template.Library()


@register.simple_tag(name='formulaic_tinymce_key')
def formulaic_tinymce_key():
    return settings.FORMULAIC_TINYMCE_KEY




import csv
import pytz
from six import u
from tzlocal import get_localzone

from formulaic import models, utils


def export_submissions_to_file(form, output_file):
    field_names = form.column_headers
    writer = csv.DictWriter(output_file, fieldnames=field_names)
    writer.writeheader()

    # Prefetch the values, with the field select_related for speed.
    submission_qs = (
        models.Submission.objects
        .filter(form=form)
        .order_by('id')
        .prefetch_custom_data()
    )

    for _, _, _, submission_batch in utils.batch_qs(submission_qs):
        row_batch = []
        for submission in submission_batch:
            row = submission.prefetched_custom_data
            local_tz = get_localzone()
            d = submission.date_created
            if d.tzinfo is None or d.tzinfo.utcoffset(d) is None:
                date_created_aware = (
                    pytz.timezone(local_tz.zone).localize(submission.date_created)
                )
            else:
                date_created_aware = submission.date_created
            row["date"] = date_created_aware.strftime('%m/%d/%Y %H:%M')
            row["source"] = submission.source
            row['promo_source'] = submission.promo_source
            row_batch.append({k: u(v) for (k, v) in row.items()})
        writer.writerows(row_batch)



from django.http import StreamingHttpResponse
from pyzipcode import ZipCodeDatabase
import us


def batch_qs(qs, batch_size=1000):
    """
    Returns a (start, end, total, queryset) tuple for each batch in the given
    queryset.

    Usage:
        # Make sure to order your querset
        article_qs = Article.objects.order_by('id')
        for start, end, total, qs in batch_qs(article_qs):
            print "Now processing %s - %s of %s" % (start + 1, end, total)
            for article in qs:
                print article.body

    Credit:
    jkocherhans
    https://djangosnippets.org/snippets/1170/
    """
    total = qs.count()
    for start in range(0, total, batch_size):
        end = min(start + batch_size, total)
        yield (start, end, total, qs[start:end])


def send_file(request, filename, full_path):
    """
    Send a file through Django without loading the whole file into
    memory at once. The FileWrapper will turn the file object into an
    iterator for chunks of 8KB.

    Credit (modified to fit our purposes):
    jcrocholl
    https://djangosnippets.org/snippets/365/
    """
    wrapper = open(full_path)
    response = StreamingHttpResponse(wrapper, content_type='text/csv')
    # response['Content-Length'] = os.path.getsize(full_path)
    response['Content-Disposition'] = 'attachment; filename="{}"'.format(filename)
    # response.set_cookie("fileDownload", value="true", max_age=60*60, path="/")
    return response


def state_from_zip(zipcode):
    try:
        zip_info = ZipCodeDatabase()[zipcode]
    except IndexError:
        return None

    if zip_info.city in ('APO', 'FPO', 'DPO'):
        # we cannot lookup state names for APO/FPO/DPO zipcodes;
        # probably we need better handling in audb for military PO addresses
        return zip_info.state

    state = us.states.lookup(zip_info.state)
    return state.name if state else None


def city_from_zip(zipcode):
    try:
        zip_info = ZipCodeDatabase()[zipcode]
        return zip_info.city
    except IndexError:
        return None



import json

from django.core.exceptions import ValidationError
from django.forms.fields import CharField, ChoiceField, HiddenInput, MultiValueField, Select
from django.forms.utils import ErrorList
from nameparser import HumanName
from six import iteritems, text_type

from formulaic.widgets import GroupedChoiceWidget


class FullNameField(CharField):
    def clean(self, value):
        cleaned_text = super(FullNameField, self).clean(value)

        parsed_name = HumanName(cleaned_text)

        return {
            None: parsed_name.full_name,
            "first": parsed_name.first,
            "last": parsed_name.last,
            "middle": parsed_name.middle,
            "suffix": parsed_name.suffix,
            "title": parsed_name.title,
        }


class GroupedChoiceField(MultiValueField):
    def __init__(self, groups={}, initial=None, field_class=ChoiceField, widget_class=Select, widget_attrs={}, *args, **kwargs):
        # create hidden_field for tracking selected field
        data_initial = initial if isinstance(initial, text_type) else json.dumps(initial)
        group_id_field = CharField(widget=HiddenInput(attrs={"data-initial": data_initial}), initial=None)

        fields = [group_id_field]
        widgets = [group_id_field.widget]
        initial_values = [group_id_field.initial]
        self.group_field_mapping = {}
        field_index = 1
        for key, options in iteritems(groups):
            attrs = {
                "data-group-id": json.dumps(key)
            }
            attrs.update(widget_attrs)

            field = field_class(
                choices=options,
                widget=widget_class(attrs=attrs)
            )
            fields.append(field)

            widgets.append(field.widget)
            initial_values.append(initial)

            self.group_field_mapping[key] = field_index
            field_index += 1

        multi_widget = GroupedChoiceWidget(widgets=widgets, widget_name=widget_class.__name__.lower())

        super(GroupedChoiceField, self).__init__(widget=multi_widget, initial=initial_values, fields=fields, *args, **kwargs)

    def compress(self, data_list):
        """
        Returns the value from the selected field and ignores the others
        """
        if data_list:
            group_id = data_list[0] or None
            if group_id is not None:
                group_id = int(group_id)
            field_index = self.group_field_mapping[group_id]

            return data_list[field_index]
        else:
            return None

    def clean(self, value):
        """
        Overriding default behavior.  Can safely ignore validation
        on all but one of the fields.
        1. Clean data received from all fields; invalid values cause
           validation to fail.
        2. If self.required, check that self.compress returns a value
        """

        clean_data = []
        errors = ErrorList()

        # ensure `value` is a list
        if value and not isinstance(value, (list, tuple)):
            raise ValidationError(self.error_messages['invalid'])

        # clean data
        for i, field in enumerate(self.fields):
            try:
                field_value = value[i]
            except IndexError:
                field_value = None

            try:
                clean_data.append(field.clean(field_value))
            except ValidationError as e:
                errors.extend(e.messages)

        """
        For now, fail validation if any field is invalid. We can
        limit to the selected field if there is a benefit.
        """
        if errors:
            raise ValidationError(errors)

        out = self.compress(clean_data)

        if self.required and not out:
            raise ValidationError(self.error_messages['required'])

        return out




from django.conf import settings
from django.forms.widgets import MultiWidget
from django.forms.widgets import TextInput


class GroupedChoiceWidget(MultiWidget):
    def __init__(self, widgets, attrs=None, widget_name="unknown"):
        self.widget_name = widget_name
        super(GroupedChoiceWidget, self).__init__(widgets, attrs=attrs)

    def decompress(self, value):
        if value and len(value) > 0:
            return [value]
        else:
            return []


class PhoneInput(TextInput):
    """Phone input field built upon Jack O'Connor's intl-tel-input widget.
    https://github.com/jackocnr/intl-tel-input
    """
    def __init__(self, attrs=None):
        super().__init__(attrs=attrs)

        self.attrs["autocomplete"] = "tel"
        self.attrs.setdefault("utilsScript", settings.STATIC_URL.rstrip("/") + "/formulaic/js/intTelInput_utils.js")
        self.attrs.setdefault("extensionPrefix", " ext. ")
        self.attrs.setdefault("fullSuffix", "_full")

    def value_from_datadict(self, data, files, name):
        """
        Instead of getting the "name" value, we want to get the "full" version.
        This version of the phone number has the country code and preformatted
        by the front end for our convenience.
        """
        suffix = self.attrs["fullSuffix"]
        full_number = data.get(name + suffix)

        # Determine the extension. This gets stripped from the "full" phone
        # number. Here we check for it, and add it on to the full number.
        formatted_number = data.get(name, "")
        ext_pre = self.attrs["extensionPrefix"]
        if ext_pre in formatted_number:
            # Find the location of the extension, slice the string and add it
            # to full_number
            full_number += formatted_number[formatted_number.find(ext_pre):]

        return full_number

    class Media:
        js = (
            "formulaic/js/intlTelInput.min.js",
            "formulaic/js/phoneNumber.js",
        )
        css = {
            "all": (
                "formulaic/css/intlTelInput.css",
            ),
        }



from datetime import date
import uuid
import os

from django.test import TestCase

from formulaic import models
from formulaic import csv_export
from formulaic.forms import CustomForm



class CsvExportTestCase(TestCase):
    def setUp(self):
        # Create Form
        form, created = models.Form.objects.get_or_create(name="My Test Form", slug="my-test-form")
        self.form_id = form.id

        # Add TextField
        models.TextField.objects.get_or_create(
            name="Text Field 1",
            slug="text-field-1",
            required=True,
            help_text=None,
            position=0,
            form_id=self.form_id,
            enabled=0, 
            css_class=None,
            subtype=models.TextField.SUBTYPE_TEXT,
        )

    def test_csv_export(self):
        # Test with 10000, 1000, 100
        number_of_submissions = 5
        file_name = "/tmp/test_formulaic_csv_export_test"
        post_data = {
            "text-field-1": "My Test Value"
        }

        form = models.Form.objects.get(pk=self.form_id)
        custom_form = CustomForm(
            post_data,
            instance_id="primary_form",
            form=form,
            label_suffix="",
            request=None
        )

        self.assertTrue(custom_form.is_valid(), "Form isn't valid.")

        # Create Submissions
        for n in range(number_of_submissions):
            form.create_submission(custom_form.cleaned_data, source=custom_form.instance_id)

        # Write CSV
        with open(file_name, "w") as outfile:
            csv_exporter = csv_export.export_submissions_to_file(form, outfile)

        # Read CSV
        with open(file_name, "r") as outfile:
            # CSV Header
            outfile.readline()
            for row in outfile:
                array = row.split(",")
                self.assertTrue(array[1], "primary_form")


from django.urls import include, re_path
from rest_framework import routers
from formulaic import views

router = routers.SimpleRouter()
router.register(r'forms', views.FormViewset)
router.register(r'privacypolicies', views.PrivacyPolicyViewset)
router.register(r'fields', views.FieldViewset)
router.register(r'textfields', views.TextFieldViewset)
router.register(r'choicefields', views.ChoiceFieldViewset)
router.register(r'booleanfields', views.BooleanFieldViewset)
router.register(r'hiddenfields', views.HiddenFieldViewset)
router.register(r'optionlists', views.OptionListViewset)
router.register(r'optiongroups', views.OptionGroupViewset)
router.register(r'options', views.OptionViewset)
router.register(r'ruleconditions', views.RuleConditionViewset)
router.register(r'rules', views.RuleViewset)
router.register(r'ruleresults', views.RuleResultViewset)
router.register(r'submissions', views.SubmissionViewset)

urlpatterns = [
    re_path(r'^api/submissionsources/$', views.SubmissionSourceView.as_view()),
    re_path(r'^api/', include(router.urls)),
    re_path(r'^download/submissions/$', views.download_submissions),
]



import uuid

from formulaic.csv_export import export_submissions_to_file
from formulaic.models import Form
from handl import errors as handl_errors
from handl.decorators import handl_job
from handl.models import DownloadableFile


@handl_job(max_retries=7)  # ~ 2 minutes
def create_submission_csv(data):

    try:
        form = Form.objects.get(pk=data.get('form_pk'))
    except Form.DoesNotExist:
        raise handl_errors.HandlTaskMissingData

    with open("/tmp/create_lead_csv-{}".format(uuid.uuid4()), "w") as outfile:
        export_submissions_to_file(form, outfile)

        files = [
            DownloadableFile(
                filename='{}-submissions.csv'.format(form.slug),
                file=outfile
            )
        ]

        return files



# Generated by Django 3.1.3 on 2020-12-07 22:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('formulaic', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='field',
            name='enabled',
            field=models.BooleanField(default=True),
        ),
    ]



import pytz
import six
from tzlocal import get_localzone

from formulaic import models
from rest_framework import serializers


class BooleanFieldSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.BooleanField
        fields = (
            'id',
            'display_name',
            'data_name',
            'slug',
            'required',
            'help_text',
            'model_class',
            'position',
            'css_class',
            'form',
            'subtype',
            'default_checked',
        )


class HiddenFieldSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.HiddenField
        fields = (
            'id',
            'display_name',
            'data_name',
            'slug',
            'required',
            'help_text',
            'model_class',
            'position',
            'css_class',
            'form',
            'subtype',
            'value',
        )


class JsonField(serializers.Field):

    def to_internal_value(self, data):
        types = [int, six.text_type, dict, list]
        if any(map(lambda t: isinstance(data, t), types)):
            return data
        else:
            msg = self.error_messages.get('invalid', "unknown error")
            raise serializers.ValidationError(msg)

    def to_representation(self, obj):
        return obj


class DefaultOptionField(serializers.Field):

    def to_internal_value(self, data):
        if isinstance(data, six.text_type):
            return data
        else:
            msg = self.error_messages['invalid']
            raise serializers.ValidationError(msg)

    def to_representation(self, obj):
        return obj


class DefaultOptionsField(serializers.Field):

    def to_internal_value(self, data):
        if isinstance(data, list):
            return data
        else:
            msg = self.error_messages['invalid']
            raise serializers.ValidationError(msg)

    def to_representation(self, obj):
        return obj


class ChoiceFieldSerializer(serializers.ModelSerializer):
    default_option = DefaultOptionField(allow_null=True)
    default_options = DefaultOptionsField(allow_null=True)

    class Meta:
        model = models.ChoiceField
        fields = (
            'id',
            'display_name',
            'data_name',
            'slug',
            'required',
            'help_text',
            'model_class',
            'position',
            'css_class',
            'form',
            #'multiselect',
            'minimum_selections',
            'maximum_selections',
            'option_list',
            'option_group',
            'default_option',
            'default_options',
            'default_text',
            'subtype',
        )


class TextFieldSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.TextField
        fields = (
            'id',
            'display_name',
            'data_name',
            'slug',
            'required',
            'help_text',
            'model_class',
            'position',
            'css_class',
            'form',
            'subtype',
        )


class FieldSerializer(serializers.ModelSerializer):
    booleanfield = BooleanFieldSerializer()
    choicefield = ChoiceFieldSerializer()
    textfield = TextFieldSerializer()
    hiddenfield = HiddenFieldSerializer()

    class Meta:
        model = models.Field
        fields = (
            'id',
            'display_name',
            'data_name',
            'slug',
            'required',
            'help_text',
            'model_class',
            'position',
            'css_class',
            'form',
            'subtype',
            #'enabled',
            "textfield",
            "booleanfield",
            "choicefield",
            "hiddenfield",
            "content_type",
        )


class FormSerializer(serializers.ModelSerializer):
    #fields = FieldSerializer(source='field_set', many=True)

    class Meta:
        model = models.Form
        fields = (
            'id',
            'name',
            'slug',
            'success_message',
            'privacy_policy',
            #'fields',
        )


class PrivacyPolicySerializer(serializers.ModelSerializer):

    class Meta:
        model = models.PrivacyPolicy
        fields = (
            'id',
            'name',
            'text',
        )


class OptionSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Option
        fields = (
            'id',
            'name',
            'value',
            'position',
            'list',
        )


class OptionListSerializer(serializers.ModelSerializer):
    options = OptionSerializer(source='option_set', many=True)

    class Meta:
        model = models.OptionList
        fields = (
            'id',
            'name',
            'options',
            'groups',
        )


class OptionGroupSerializer(serializers.ModelSerializer):
    options = OptionSerializer(many=True)

    class Meta:
        model = models.OptionGroup
        fields = (
            'id',
            'name',
            'position',
            'options',
            'list',
        )


class RuleResultSerializer(serializers.ModelSerializer):

    id = serializers.IntegerField(read_only=False, required=False)

    class Meta:
        model = models.RuleResult
        fields = (
            'id',
            'action',
            'field',
            'rule',
            'option_group',
        )


class RuleConditionSerializer(serializers.ModelSerializer):

    # id: override read_only default
    id = serializers.IntegerField(read_only=False, required=False)

    # value: handle property (not a Django field)
    value = JsonField(required=False, allow_null=True)

    class Meta:
        model = models.RuleCondition
        fields = (
            'id',
            'position',
            'rule',
            'field',
            'operator',
            'value',
        )


class RuleSerializer(serializers.ModelSerializer):
    conditions = RuleConditionSerializer(many=True)
    results = RuleResultSerializer(many=True)

    class Meta:
        model = models.Rule
        fields = (
            'id',
            'form',
            'operator',
            'position',
            'conditions',
            'results',
        )

    def update(self, rule_obj, validated_data):
        conditions = validated_data.pop('conditions')
        results = validated_data.pop('results')

        # Update rule
        rule_obj.operator = validated_data.get('operator', rule_obj.operator)
        rule_obj.position = validated_data.get('position', rule_obj.position)
        rule_obj.save()

        # Update, create, or delete conditions
        condition_objs = list(rule_obj.conditions.all())
        for condition in conditions:
            if condition.get('id'):
                # exists
                for condition_obj in condition_objs:
                    if condition.get('id') == condition_obj.id:
                        # update existing condition
                        condition_obj.position = condition.get('position', condition_obj.position)
                        condition_obj.field_id = condition.get('field', condition_obj.field_id)
                        condition_obj.operator = condition.get('operator', condition_obj.operator)
                        condition_obj.value = condition.get('value', condition_obj.value)

                        # mark to avoid deletion
                        condition_obj.keep_alive = True
            else:
                # create new condition
                condition['rule'] = rule_obj
                # condition['value'] = "Test"  # TODO: remove hard coded value

                value = condition.pop('value')

                # Note: can't use `create` method; need access to `data` property
                condition_obj = models.RuleCondition(**condition)
                condition_obj.value = value
                condition_obj.save()

        # Carry out saves / deletes
        for condition_obj in condition_objs:
            if getattr(condition_obj, 'keep_alive', False):
                condition_obj.save()
            else:
                condition_obj.delete()

        # Update, create, or delete results
        result_objs = list(rule_obj.results.all())
        for result in results:
            if result.get('id'):
                # exists
                for result_obj in result_objs:
                    if result.get('id') == result_obj.id:
                        # update existing result
                        result_obj.action = result.get('action', result_obj.action)
                        result_obj.field_id = result.get('field', result_obj.field_id)
                        result_obj.option_group = result.get('option_group', result_obj.option_group)

                        # mark to avoid deletion
                        result_obj.keep_alive = True
            else:
                # create new rule
                result['rule'] = rule_obj
                models.RuleResult.objects.create(**result)

        # Carry out saves / deletes
        for result_obj in result_objs:
            if getattr(result_obj, 'keep_alive', False):
                result_obj.save()
            else:
                result_obj.delete()

        return rule_obj

    def create(self, validated_data):
        conditions = validated_data.pop('conditions')
        results = validated_data.pop('results')

        # Create rule
        rule = models.Rule.objects.create(**validated_data)

        # Create all conditions
        for condition in conditions:
            condition['rule'] = rule
            # condition['value'] = "Test"  # TODO: remove hard coded value

            value = condition.pop('value', None)

            # Note: can't use `create` method; need access to `data` property
            condition_obj = models.RuleCondition(**condition)
            condition_obj.value = value
            condition_obj.save()

        # Create all results
        for result in results:
            result['rule'] = rule
            models.RuleResult.objects.create(**result)

        return rule


class CustomDateTimeField(serializers.DateTimeField):

    def to_representation(self, obj):
        try:
            local_tz = get_localzone()
            obj_aware = pytz.timezone(local_tz.zone).localize(obj)
        except ValueError:
            obj_aware = obj

        return obj_aware.strftime('%m/%d/%Y %H:%M %Z')


class SubmissionSerializer(serializers.ModelSerializer):
    # values: handle property (not a Django field)
    custom_data = serializers.ReadOnlyField()

    date_created = CustomDateTimeField()

    class Meta:
        model = models.Submission
        fields = (
            'id',
            'date_created',
            'form',
            'custom_data',
            'source',
            'promo_source',
        )



from django.contrib.auth import get_user_model
from django.test import TestCase
from django.urls import reverse
from rest_framework.test import APIClient

from formulaic import models


class FieldViewsetTestCase(TestCase):

    def setUp(self):
        """Construct a collection of objects that will frequently get used.

        self.field_1 relates to a field that is associated with a RuleCondition
        of a Rule. self.field_2 relates to a RuleResult of the same Rule.
        """
        User = get_user_model()
        self.superuser = User.objects.create_superuser('superuser', 'superuser@test.com', "password")
        self.client = APIClient()

        # create form
        self.form = models.Form.objects.create(name="My Test Form", slug="my-test-form")

        # add textfield
        self.field_1 = models.TextField.objects.create(
            name="Text Field 1",
            slug="text-field-1",
            required=True,
            data_name="text_field_1",
            position=0,
            form=self.form,
            subtype=models.TextField.SUBTYPE_TEXT,
        )

        self.field_2 = models.TextField.objects.create(
            name="Text Field 2",
            slug="text-field-2",
            required=True,
            data_name="text_field_2",
            position=1,
            form=self.form,
            subtype=models.TextField.SUBTYPE_TEXT,
        )

        self.rule = models.Rule.objects.create(
            form=self.form,
            position=1,
            operator=models.Rule.OPERATOR_AND
        )
        self.condition = models.RuleCondition.objects.create(
            rule=self.rule,
            field=self.field_1,
            position=1,
            operator=models.RuleCondition.OPERATOR_CONTAINS,
            value_string="testing"
        )

        self.result = models.RuleResult.objects.create(
            rule=self.rule,
            field=self.field_2,
            action=models.RuleResult.ACTION_SHOW
        )

    def test_destroy_last_condition(self):
        """
        Confirm that if you delete a field that is the linked to the
        last condition for a given Rule, the Field, RuleCondition, RuleResult
        and Rule are all deleted.
        """
        # Make the request to destroy.
        url = reverse('textfield-detail', kwargs={'pk': self.field_1.pk})
        self.client.force_authenticate(user=self.superuser)
        resp = self.client.delete(url)
        self.assertEqual(resp.status_code, 204)

        # Confirm that field_1, rule, condition, result are all deleted.
        self.assertFalse(models.TextField.objects.filter(pk=self.field_1.pk).exists())
        self.assertFalse(models.Rule.objects.filter(pk=self.rule.pk).exists())
        self.assertFalse(models.RuleCondition.objects.filter(pk=self.condition.pk).exists())
        self.assertFalse(models.RuleResult.objects.filter(pk=self.result.pk).exists())

        # Confirm that field_2 still exists.
        self.assertTrue(models.TextField.objects.filter(pk=self.field_2.pk).exists())

    def test_destroy_not_last_condition(self):
        """
        Confirm that if you delete a field that is the linked to the
        condition of a Rule, but it is not the only condition.

        In this scenario we expect the Condition related to the field to be
        deleted, but the Rule and RuleResult should remain.
        """

        # Setup extra condition so that self.condition is not the last condition.
        field_3 = models.TextField.objects.create(
            name="Text Field 3",
            slug="text-field-3",
            required=True,
            data_name="text_field_3",
            position=1,
            form=self.form,
            subtype=models.TextField.SUBTYPE_TEXT,
        )

        extra_condition = models.RuleCondition.objects.create(
            rule=self.rule,
            field=field_3,
            position=2,
            operator=models.RuleCondition.OPERATOR_CONTAINS,
            value_string="testing"
        )

        # Make the request to destroy.
        url = reverse('textfield-detail', kwargs={'pk': self.field_1.pk})
        self.client.force_authenticate(user=self.superuser)
        resp = self.client.delete(url)
        self.assertEqual(resp.status_code, 204)

        # Confirm that only field_1 and the related condition is deleted.
        self.assertFalse(models.TextField.objects.filter(pk=self.field_1.pk).exists())
        self.assertFalse(models.RuleCondition.objects.filter(pk=self.condition.pk).exists())

        # Confirm that field_2, the new condition and result still exist.
        self.assertTrue(models.TextField.objects.filter(pk=self.field_2.pk).exists())
        self.assertTrue(models.Rule.objects.filter(pk=self.rule.pk).exists())
        self.assertTrue(models.RuleCondition.objects.filter(pk=extra_condition.pk).exists())
        self.assertTrue(models.RuleResult.objects.filter(pk=self.result.pk).exists())

    def test_destroy_last_result(self):
        """
        Confirm that if you delete a field that is the linked to the
        last condition for a given Rule, the Field, RuleCondition, RuleResult
        and Rule are all deleted.
        """
        # Make the request to destroy.
        url = reverse('textfield-detail', kwargs={'pk': self.field_2.pk})
        self.client.force_authenticate(user=self.superuser)
        resp = self.client.delete(url)
        self.assertEqual(resp.status_code, 204)

        # Confirm that field_2, rule, condition, result are all deleted.
        self.assertFalse(models.TextField.objects.filter(pk=self.field_2.pk).exists())
        self.assertFalse(models.Rule.objects.filter(pk=self.rule.pk).exists())
        self.assertFalse(models.RuleCondition.objects.filter(pk=self.condition.pk).exists())
        self.assertFalse(models.RuleResult.objects.filter(pk=self.result.pk).exists())

        # Confirm that field_1 still exists.
        self.assertTrue(models.TextField.objects.filter(pk=self.field_1.pk).exists())

    def test_destroy_not_last_result(self):
        """
        Confirm that if you delete a field that is the linked to the
        result of a Rule, but it is not the only rule.

        In this scenario we expect the Rule related to the field to be
        deleted, but the Rule and Condition should remain.
        """

        # Setup extra condition so that self.condition is not the last condition.
        field_3 = models.TextField.objects.create(
            name="Text Field 3",
            slug="text-field-3",
            required=True,
            data_name="text_field_3",
            position=1,
            form=self.form,
            subtype=models.TextField.SUBTYPE_TEXT,
        )

        extra_result = models.RuleResult.objects.create(
            rule=self.rule,
            field=field_3,
            action=models.RuleResult.ACTION_SHOW
        )

        # Make the request to destroy.
        url = reverse('textfield-detail', kwargs={'pk': self.field_2.pk})
        self.client.force_authenticate(user=self.superuser)
        resp = self.client.delete(url)
        self.assertEqual(resp.status_code, 204)

        # Confirm that only field_2 and the related rule is deleted.
        self.assertFalse(models.TextField.objects.filter(pk=self.field_2.pk).exists())
        self.assertFalse(models.RuleResult.objects.filter(pk=self.result.pk).exists())

        # Confirm that field_1, the new result and condition still exist.
        self.assertTrue(models.TextField.objects.filter(pk=self.field_1.pk).exists())
        self.assertTrue(models.Rule.objects.filter(pk=self.rule.pk).exists())
        self.assertTrue(models.RuleResult.objects.filter(pk=extra_result.pk).exists())
        self.assertTrue(models.RuleCondition.objects.filter(pk=self.condition.pk).exists())



from bs4 import BeautifulSoup
from django import forms
import phonenumbers


def validate_mixed_content(value):
    """
    Validate content to avoid mixed content warnings
    """

    targets = (
        {"tag": "img", "attr": "src"},
        {"tag": "script", "attr": "src"},
        {"tag": "iframe", "attr": "src"},
        {"tag": "link", "attr": "src"},
    )

    soup = BeautifulSoup(value, 'html.parser')
    errors = []

    for target in targets:
        for item in soup.find_all(target["tag"]):
            src = item.get(target["attr"], '')
            if "http://" in src:
                if not errors:
                    errors.append("These tags must use https protocol:")
                errors.append("<{}>: {}".format(target["tag"], src))

    if errors:
        raise forms.ValidationError(errors)


def validate_phone_number(value):
    """Uses the phonenumbers library to try and parse the phone number and
    check for it's validity. """

    try:
        z = phonenumbers.parse(value, None)
    except phonenumbers.NumberParseException:
        raise forms.ValidationError("Enter a valid phone number.")

    if not phonenumbers.is_valid_number(z):
        raise forms.ValidationError("Enter a valid phone number.")




# Generated by Django 3.1.4 on 2021-03-12 20:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('formulaic', '0003_auto_20201230_1406'),
    ]

    operations = [
        migrations.AddField(
            model_name='submission',
            name='promo_source',
            field=models.CharField(blank=True, help_text='Source passed through metadata variable promo_source', max_length=200, null=True),
        ),
    ]



from django.test import TestCase

from formulaic.validators import validate_phone_number
from django.forms import ValidationError


class PhoneNumberTestCase(TestCase):

    def test_valid_phone_number(self):
        """Confirm that no exception gets raised by this function when provided
        with a "valid" phone number"""
        self.assertIsNone(validate_phone_number("+12025551234"))

    def test_valid_phone_number_with_extension(self):
        """Confirm that no exception gets raised by this function when provided
        with a "valid" phone number"""
        self.assertIsNone(validate_phone_number("+12025551234 ext. 12"))

    def test_possible_but_invalid_phone_number(self):
        """These are numbers that look like they could be real, but are not.
        An example would be one that uses an area code that does not exist."""

        with self.assertRaises(ValidationError):
            validate_phone_number("+19995551234")

    def test_invalid_phone_number(self):
        """Other junky phone numbers."""

        # Missing one digit.
        with self.assertRaises(ValidationError):
            validate_phone_number("+1202555123")

        # Too many digits
        with self.assertRaises(ValidationError):
            validate_phone_number("+120255512345")

        # Missing Country Code
        with self.assertRaises(ValidationError):
            validate_phone_number("2025551234")

        # Whatever
        with self.assertRaises(ValidationError):
            validate_phone_number("Whatever")



from six import iteritems, u

from formulaic.utils import state_from_zip, city_from_zip


"""
#TODO#
This way of auto-populating state values from zipcode values is a work-around until
we can add a proper `ZipCodeField` field to Formulaic, and then implement the
auto-zipcode-to-state mapping in a way similar to how it was done for the `FullNameField`
parsing. Right now we are not doing the more correct thing for time-saving reasons.
"""


def attempt_kv_auto_populate(submission_kv, field_val, form_data):
    """
    Attempts to auto-populate values for certain form fields; passes through the original
    value (which in most cases will be an empty/null of some kind; why would we be auto-
    populating otherwise?) untouched if no auto-population was possible.
    """
    if submission_kv.key.lower() == "state" or submission_kv.key.lower() == "city":
        return _attempt_state_or_city_from_zipcode(submission_kv, field_val, form_data)

    # . . . other kinds of auto-populate attempts could be added here
    return field_val


def _attempt_state_or_city_from_zipcode(submission_kv, field_val, form_data):
    """
    Attempt to get the state name from the zipcode, otherwise pass the original value
    through.

    We only look for hidden 'state' fields, since that field configuration is the user's
    way of signalling that they want 'state' to be auto-populated by Formulaic. It's
    certainly a not-ideal work-around.
    """
    is_hidden_field = submission_kv.field.subtype_is(submission_kv.field.TYPE_HIDDEN)

    zipcode = None
    for key, value in iteritems(form_data):
        if key.lower() not in ("zipcode", "zip-code"):
            continue
        try:
            zipcode = value.strip()[:5]
        except AttributeError:
            zipcode = u(value).strip()[:5]
        break

    if is_hidden_field and zipcode and not field_val:
        if submission_kv.key.lower() == "state":
            return state_from_zip(zipcode)
        else:
            return city_from_zip(zipcode)

    return field_val



from six import iteritems

from formulaic import exceptions


class RuleAssessor(object):
    """
    Mimics the Rule implementation found in JavaScript to toggle
    validation on and off for fields depending on if they're
    supposed to be visible to the user.
    """

    def __init__(self, rules_data, form_fields, submission_data={}):
        self.fields = {}
        for slug, form_field in form_fields.items():
            value = submission_data.get(slug, None)

            self.fields[slug] = Field(slug, form_field, value)

        self.rules = []
        for rule_data in rules_data:
            self.rules.append(Rule(self, rule_data))

        self.visible_fields = []
        self.invisible_fields = []
        for slug, field in iteritems(self.fields):
            field.evaluate_observed_rules()

            if field.visible:
                self.visible_fields.append(slug)
            else:
                self.invisible_fields.append(slug)

    def is_field_visible(self, slug):
        return slug in self.visible_fields


class FieldStatus(object):
    def __init__(self):
        self.visible = True


class Field(object):
    def __init__(self, slug, form_field, value):
        self.slug = slug
        self.form_field = form_field
        self.value = value
        self.observed_rules = []

        self._visible = None

    def add_observed_rule(self, rule):
        self.observed_rules.append(rule)

    def evaluate_observed_rules(self):
        new_field_status = FieldStatus()

        for rule in self.observed_rules:
            rule.evaluate(self, new_field_status)

        # apply field status
        self._visible = new_field_status.visible

    def has_value(self, value):
        from django.forms import CheckboxInput
        if type(self.form_field) is CheckboxInput:
            return self.value
        elif type(self.value) is list:
            return value in self.value
        else:
            return self.value == value

    @property
    def visible(self):
        if self._visible is None:
            raise exceptions.NotSetException("Field hasn't been set active/inactive, yet")
        else:
            return self._visible


class Rule(object):
    OPERATOR_AND = "and"
    OPERATOR_OR = "or"

    def __init__(self, assessor, rule_data):
        self.assessor = assessor
        self.operator = rule_data["operator"]

        self.conditions = []
        for condition_data in rule_data["conditions"]:
            self.conditions.append(RuleCondition(self, condition_data))

        self.results = []
        for result_data in rule_data["results"]:
            self.results.append(RuleResult(self, result_data))

    def conditions_met(self):
        require_all = self.operator == Rule.OPERATOR_AND
        all_true = True
        any_true = False

        for condition in self.conditions:
            if condition.is_met():
                any_true = True
            else:
                all_true = False

        if all_true:
            return True
        elif not require_all and any_true:
            return True
        else:
            return False

    def evaluate(self, field, display_status):
        if self.conditions_met():
            for result in self.results:
                result.update_result(field, display_status)


class RuleCondition(object):
    def __init__(self, rule, condition_data):
        self.rule = rule

        self.field = self.rule.assessor.fields[condition_data["field_slug"]]
        self.operator = condition_data["operator"]
        self.value = condition_data["value"]

    def is_met(self):
        if self.operator == "is":
            return self.field.has_value(self.value)
        else:
            return not self.field.has_value(self.value)


class RuleResult(object):
    ACTION_SHOW = "show"
    ACTION_HIDE = "hide"
    ACTION_CHANGE_OPTION_GROUP = "change-option-group"

    def __init__(self, rule, result_data):
        self.rule = rule

        self.action = result_data["action"]
        self.field = self.rule.assessor.fields[result_data["field_slug"]]

        self.field.add_observed_rule(self.rule)

    def update_result(self, field, display_status):
        if self.field == field:
            if self.action == RuleResult.ACTION_SHOW:
                display_status.visible = True
            elif self.action == RuleResult.ACTION_HIDE:
                display_status.visible = False
            elif self.action == RuleResult.ACTION_CHANGE_OPTION_GROUP:
                pass
            else:
                raise exceptions.InvalidParamException("Invalid RuleResult action: {}".format(self.action))



from django.test import TestCase

from formulaic import models
from formulaic.forms import CustomForm


class SubmissionTestCase(TestCase):
    def setUp(self):
        # create form
        form, created = models.Form.objects.get_or_create(
            name="My Test Form", slug="my-test-form"
        )
        self.form_id = form.id

        # add textfield
        models.TextField.objects.get_or_create(
            name="Text Field 1",
            slug="text-field-1",
            required=True,
            help_text=None,
            position=0,
            form_id=self.form_id,
            enabled=0,
            css_class=None,
            subtype=models.TextField.SUBTYPE_TEXT,
        )

        # add phone number
        models.TextField.objects.get_or_create(
            name="Phone Number 1",
            slug="phone-number-1",
            required=True,
            help_text=None,
            position=0,
            form_id=self.form_id,
            enabled=0,
            css_class=None,
            subtype=models.TextField.SUBTYPE_PHONE_NUMBER,
        )

    def test_submission_with_promo_source_saves_properly(self):
        post_data = {
            "text-field-1": "My Test Value",
            "phone-number-1": "(202) 555 1234",
            "phone-number-1_full": "+12025551234"
        }

        form = models.Form.objects.get(pk=self.form_id)
        custom_form = CustomForm(
            post_data,
            instance_id="primary_form",
            form=form,
            label_suffix="",
            request=None,
        )
        custom_form.is_valid()
        self.assertTrue(custom_form.is_valid(), "Form isn't valid.")

        obj = form.create_submission(
            custom_form.cleaned_data,
            source=custom_form.instance_id,
            promo_source="testing_promo_source",
        )

        submission = models.Submission.objects.get(pk=obj.pk)
        self.assertEqual(submission.promo_source, "testing_promo_source")



from django import template

register = template.Library()


def autocomplete_attribute(name):
    attributes = {
        "email": "email",
        "full_name": "name",
        "full-name": "name",
        "custom_companyname": "organization",
        "job-title": "organization-title",
        "phone": "tel",
        "phone-number": "tel",
        "zip-code": "postal-code",
        "country_other": "country-name",
    }

    return attributes[name] if name in attributes else "on"


def widget_type(field):
    """
    (stolen from django-widget-tweaks) Returns field widget class name (in lower case).
    """
    if hasattr(field, 'field') and hasattr(field.field, 'widget') and field.field.widget:
        widget_name = field.field.widget.__class__.__name__.lower()
        if widget_name == "groupedchoicewidget":
            widget_name = field.field.widget.widget_name

        return widget_name
    return ''


def add_attributes_to_field(field):
    """ Adds extra attributes to the form widget """
    if field.field.required:
        field.field.widget.attrs["data-required"] = "true"

    if widget_type(field) == "textinput" and field.name:
        field.field.widget.attrs.update({
            "autocomplete": autocomplete_attribute(field.name)
        })

    return field


def add_classes_to_field(field, extra_classes):
    """ Adds extra CSS classes to the form widget """
    classes = field.field.widget.attrs.get("class", "").split()
    classes.extend(extra_classes.split())
    field.field.widget.attrs["class"] = " ".join(sorted(set(classes)))
    return field


def formulaic_field_classes(widget):
    classes = []

    # type of widget
    widget_name = widget.__class__.__name__.lower()

    if widget_name == "groupedchoicewidget":
        widget_name = widget.widget_name
    classes.append(widget_name)

    # hide/show
    readonly = widget.attrs.get("readonly", False)
    disabled = widget.attrs.get("disabled", False)
    if readonly or disabled:
        classes.append("disabled")

    return " ".join(classes)


register.filter('formulaic_extra_attributes', add_attributes_to_field)
register.filter('formulaic_extra_widget_classes', add_classes_to_field)
register.filter('formulaic_field_classes', formulaic_field_classes)
register.filter('widget_type', widget_type)



from datetime import datetime

from django.contrib.auth.decorators import permission_required
from django.db.models import Count
from django.db import transaction
from django.http import Http404
from django.views.decorators.cache import never_cache
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import permissions, viewsets, pagination, views as rf_views
from rest_framework.response import Response

from formulaic import models, utils, serializers, settings
from formulaic.csv_export import export_submissions_to_file


class CustomDjangoModelPermissions(permissions.DjangoModelPermissions):
    perms_map = {
        'GET': ['%(app_label)s.change_%(model_name)s'],
        'OPTIONS': [],
        'HEAD': [],
        'POST': ['%(app_label)s.add_%(model_name)s'],
        'PUT': ['%(app_label)s.change_%(model_name)s'],
        'PATCH': ['%(app_label)s.change_%(model_name)s'],
        'DELETE': ['%(app_label)s.delete_%(model_name)s'],
    }


@permission_required("formulaic.change_submission")
@never_cache
def download_submissions(request):
    # TODO: auto-cleanup files

    form_id = request.GET.get('form', None)

    if not form_id:
        raise Http404()

    try:
        form = models.Form.objects.get(pk=form_id)
    except models.Form.DoesNotExist:
        raise Http404()

    datetime_slug = datetime.now().strftime("%Y%m%d-%H:%M:%S-%f")
    filename = '{}-submissions-{}.csv'.format(form.slug, datetime_slug)
    full_path = '{}/{}'.format(settings.FORMULAIC_EXPORT_STORAGE_LOCATION, filename)

    with open(full_path, 'w') as csvfile:
        export_submissions_to_file(form, csvfile)

    return utils.send_file(request, filename, full_path)


class SubmissionSourceView(rf_views.APIView):
    permission_classes = (CustomDjangoModelPermissions,)
    queryset = models.Submission.objects.none()  # Required for DjangoModelPermissions

    def get(self, request):
        """
        Return all unique submission sources for form
        """
        form_id = request.GET.get("form", None)

        if not form_id:
            raise Http404()

        data = (
            models.Submission.objects
            .values("source")
            .filter(form_id=form_id)
            .annotate(count=Count("source"))
            .order_by("source")
        )

        return Response(data)


class StandardResultsSetPagination(pagination.PageNumberPagination):
    page_size = 5
    page_size_query_param = 'page_size'
    max_page_size = 1000


class SubmissionViewset(viewsets.ReadOnlyModelViewSet):
    permission_classes = (
        CustomDjangoModelPermissions,
    )

    queryset = (
        models.Submission.objects
        .all()
        .order_by('-date_created')
        .prefetch_related('values')
    )

    serializer_class = serializers.SubmissionSerializer
    pagination_class = StandardResultsSetPagination

    filter_backends = [DjangoFilterBackend]
    filterset_fields = ('form', 'source',)


class RuleConditionViewset(viewsets.ModelViewSet):
    permission_classes = (
        CustomDjangoModelPermissions,
    )

    queryset = (models.RuleCondition.objects.all())

    serializer_class = serializers.RuleConditionSerializer


class RuleViewset(viewsets.ModelViewSet):
    permission_classes = (
        CustomDjangoModelPermissions,
    )

    queryset = (models.Rule.objects.all())

    serializer_class = serializers.RuleSerializer

    filter_backends = [DjangoFilterBackend]
    filterset_fields = ('form', 'position',)  # todo:remove position


class RuleResultViewset(viewsets.ModelViewSet):
    permission_classes = (
        CustomDjangoModelPermissions,
    )

    queryset = (models.RuleResult.objects.all())

    serializer_class = serializers.RuleResultSerializer


class FormViewset(viewsets.ModelViewSet):
    permission_classes = (
        CustomDjangoModelPermissions,
    )

    queryset = (models.Form.objects.all())

    serializer_class = serializers.FormSerializer


class PrivacyPolicyViewset(viewsets.ModelViewSet):
    permission_classes = (
        CustomDjangoModelPermissions,
    )

    queryset = (models.PrivacyPolicy.objects.all())

    serializer_class = serializers.PrivacyPolicySerializer


class FieldViewset(viewsets.ModelViewSet):
    permission_classes = (
        CustomDjangoModelPermissions,
    )

    queryset = (
        models.Field.objects.all().select_related("textfield", "booleanfield", "choicefield", "content_type")
    )
    serializer_class = serializers.FieldSerializer

    filter_backends = [DjangoFilterBackend]
    filterset_fields = ('form',)

    def destroy(self, request, *args, **kwargs):
        """Clean up any rules that may be associated with the object.

        # todo
        This is intended to only be a temporary change to help alleviate
        frustrations around not being able to delete fields. The ideal solution
        would be to display to the user why they cannot delete a particular field.
        At that point we can confirm the blanket delete (what this function is doing)
        or help guide the user to carefully remove the ruleresults and ruleconditions.

        If a Rule does not have a rule condition or rule result, then it is invalid. If
        a field is associated with the last rule condition or rule result, we'll also
        delete the rule.
        """
        instance = self.get_object()

        rule_ids_to_be_deleted = set()
        rule_result_ids_to_be_delete = set()
        rule_conditions_ids_to_be_delete = set()

        # Gathering rule result related deletes.
        related_rule_results = instance.ruleresult_set.all().select_related("rule").prefetch_related("rule__results")
        for rule_result in related_rule_results:

            if rule_result.rule.results.all().count() == 1:
                rule_ids_to_be_deleted.add(rule_result.rule_id)

            rule_result_ids_to_be_delete.add(rule_result.id)

        # Gathering rule condition related deletes.
        related_ruleconditions = instance.rulecondition_set.all().select_related("rule").prefetch_related("rule__conditions")
        for rule_condition in related_ruleconditions:
            if rule_condition.rule.conditions.all().count() == 1:
                rule_ids_to_be_deleted.add(rule_condition.rule_id)

            rule_conditions_ids_to_be_delete.add(rule_condition.id)

        # Actually perform the deletes. Perform it in an atomic block to help revert
        with transaction.atomic():
            models.Rule.objects.filter(id__in=rule_ids_to_be_deleted).delete()
            models.RuleResult.objects.filter(id__in=rule_result_ids_to_be_delete).delete()
            models.RuleCondition.objects.filter(id__in=rule_conditions_ids_to_be_delete).delete()
            return super().destroy(request, *args, **kwargs)


class TextFieldViewset(FieldViewset):
    permission_classes = (
        CustomDjangoModelPermissions,
    )

    queryset = (
        models.TextField.objects.all()
    )
    serializer_class = serializers.TextFieldSerializer

    filter_backends = [DjangoFilterBackend]
    filterset_fields = ('form',)


class ChoiceFieldViewset(FieldViewset):
    permission_classes = (
        CustomDjangoModelPermissions,
    )

    queryset = (
        models.ChoiceField.objects.all()
    )
    serializer_class = serializers.ChoiceFieldSerializer

    filter_backends = [DjangoFilterBackend]
    filterset_fields = ('form',)


class BooleanFieldViewset(FieldViewset):
    permission_classes = (
        CustomDjangoModelPermissions,
    )

    queryset = (
        models.BooleanField.objects.all()
    )
    serializer_class = serializers.BooleanFieldSerializer

    filter_backends = [DjangoFilterBackend]
    filterset_fields = ('form',)


class HiddenFieldViewset(FieldViewset):
    permission_classes = (
        CustomDjangoModelPermissions,
    )

    queryset = (
        models.HiddenField.objects.all()
    )
    serializer_class = serializers.HiddenFieldSerializer

    filter_backends = [DjangoFilterBackend]
    filterset_fields = ('form',)


class OptionListViewset(viewsets.ModelViewSet):
    permission_classes = (
        CustomDjangoModelPermissions,
    )

    queryset = (
        models.OptionList.objects.all()
    )
    serializer_class = serializers.OptionListSerializer


class OptionGroupViewset(viewsets.ModelViewSet):
    permission_classes = (
        CustomDjangoModelPermissions,
    )

    queryset = (
        models.OptionGroup.objects.all()
    )
    serializer_class = serializers.OptionGroupSerializer

    filter_backends = [DjangoFilterBackend]
    filterset_fields = ('list',)


class OptionViewset(viewsets.ModelViewSet):
    permission_classes = (
        CustomDjangoModelPermissions,
    )

    queryset = (
        models.Option.objects.all()
    )
    serializer_class = serializers.OptionSerializer



from __future__ import unicode_literals

import json

from django import forms
from django.template.context_processors import csrf
from django.template.loader import render_to_string
from django.utils.functional import cached_property
from django.utils.safestring import mark_safe

from formulaic.rules import RuleAssessor


class CustomForm(forms.Form):
    required_css_class = "required"

    def __init__(self, data=None, *args, **kwargs):
        self.instance_id = kwargs.pop("instance_id")
        self.template = kwargs.pop("template", "formulaic/bootstrap-form.html")

        self.request = kwargs.pop("request")
        form = kwargs.pop("form")
        widget_attrs = kwargs.pop("widget_attrs", {})
        fields = form.field_set.all()
        self.field_slugs_by_id = {}
        self.privacy_policy = form.privacy_policy

        super(CustomForm, self).__init__(data, *args, **kwargs)

        # Add fields to form
        for i, field in enumerate(fields):
            specific_field = field.get_specific_field()

            self.fields[field.slug] = specific_field.get_implementation(
                widget_attrs=widget_attrs
            )

            self.field_slugs_by_id[field.id] = field.slug

        # Add rules to form
        self._rules = form.rule_set.all()

    def _clean_fields(self):
        super(CustomForm, self)._clean_fields()

        rule_assessor = RuleAssessor(self.rules_data, self.fields, self.cleaned_data)

        for invisible_field in rule_assessor.invisible_fields:
            if invisible_field in self._errors:
                del self._errors[invisible_field]

            if invisible_field in self.cleaned_data:
                del self.cleaned_data[invisible_field]

    @cached_property
    def rules_data(self):
        rules_list = []

        for rule in self._rules:
            # convert rule's conditions
            conditions_list = []
            for condition in rule.conditions.all():
                conditions_list.append({
                    "field_id": condition.field_id,
                    "field_slug": self.field_slugs_by_id[condition.field_id],
                    "operator": condition.operator,
                    "value": condition.value,
                })

            # convert rule's results
            results_list = []
            for result in rule.results.all():
                results_list.append({
                    "field_id": result.field_id,
                    "field_slug": self.field_slugs_by_id[result.field_id],
                    "action": result.action,
                    "option_group_id": result.option_group_id,
                })

            # convert rule
            rules_list.append({
                "operator": rule.operator,
                "conditions": conditions_list,
                "results": results_list,
            })

        return rules_list

    @property
    def rules_json(self):
        return json.dumps(self.rules_data)

    def render(self):
        c = {"form": self}
        c.update(csrf(self.request))

        return mark_safe(render_to_string(self.template, c))

    def __str__(self):
        return self.render()

    class Media:
        # TODO: handle caching; can't currently use querystring because of
        # django internals
        js = ("formulaic/js/custom_form.js", )
        css = {
            "all": ("formulaic/css/custom_form.css", ),
        }



# Generated by Django 3.1.4 on 2020-12-30 19:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('formulaic', '0002_add_default_to_field_enabled'),
    ]

    operations = [
        migrations.AlterField(
            model_name='option',
            name='position',
            field=models.PositiveIntegerField(default=0, verbose_name='Position'),
        ),
        migrations.AlterField(
            model_name='optiongroup',
            name='position',
            field=models.PositiveIntegerField(default=0, verbose_name='Position'),
        ),
    ]



# Generated by Django 3.1.2 on 2020-10-28 19:32

import ckeditor.fields
from django.db import migrations, models
import django.db.models.deletion
import formulaic.validators


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Field',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=500)),
                ('display_name', models.CharField(max_length=1000)),
                ('data_name', models.CharField(max_length=200)),
                ('slug', models.SlugField(max_length=200)),
                ('required', models.BooleanField()),
                ('help_text', models.CharField(blank=True, max_length=500, null=True)),
                ('subtype', models.CharField(max_length=100)),
                ('model_class', models.CharField(max_length=100)),
                ('position', models.PositiveIntegerField(verbose_name='Position')),
                ('enabled', models.BooleanField()),
                ('css_class', models.CharField(blank=True, max_length=120, null=True)),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='contenttypes.ContentType')),
            ],
            options={
                'ordering': ['position'],
            },
        ),
        migrations.CreateModel(
            name='Form',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=500)),
                ('slug', models.SlugField(max_length=200)),
                ('success_message', models.TextField(blank=True, null=True)),
                ('archived', models.BooleanField()),
            ],
            options={
                'ordering': ('archived', 'name'),
            },
        ),
        migrations.CreateModel(
            name='Option',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=250)),
                ('value', models.CharField(max_length=250)),
                ('position', models.PositiveIntegerField(verbose_name='Position')),
            ],
            options={
                'ordering': ('position',),
            },
        ),
        migrations.CreateModel(
            name='OptionGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=250)),
                ('position', models.PositiveIntegerField(verbose_name='Position')),
            ],
            options={
                'ordering': ('position',),
            },
        ),
        migrations.CreateModel(
            name='OptionList',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=250)),
            ],
        ),
        migrations.CreateModel(
            name='PrivacyPolicy',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=250)),
                ('text', ckeditor.fields.RichTextField(validators=[formulaic.validators.validate_mixed_content])),
            ],
            options={
                'verbose_name_plural': 'Privacy policies',
            },
        ),
        migrations.CreateModel(
            name='Rule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('operator', models.CharField(choices=[('and', 'And'), ('or', 'Or')], max_length=3)),
                ('position', models.IntegerField()),
                ('form', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='formulaic.Form')),
            ],
            options={
                'ordering': ('position',),
            },
        ),
        migrations.CreateModel(
            name='Submission',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('date_created', models.DateTimeField()),
                ('source', models.CharField(blank=True, help_text='Name of the specific form placement the user filled out (e.g. sidebar-subscribe)', max_length=200, null=True)),
                ('metadata_serialized', models.TextField(help_text='Serialized JSON object storing arbitrary submission-related metadata')),
                ('form', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='formulaic.Form')),
            ],
        ),
        migrations.CreateModel(
            name='BooleanField',
            fields=[
                ('field_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='formulaic.Field')),
                ('default_checked', models.BooleanField()),
            ],
            bases=('formulaic.Field',),
        ),
        migrations.CreateModel(
            name='HiddenField',
            fields=[
                ('field_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='formulaic.Field')),
                ('value', models.CharField(blank=True, max_length=500, null=True)),
            ],
            bases=('formulaic.Field',),
        ),
        migrations.CreateModel(
            name='TextField',
            fields=[
                ('field_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='formulaic.Field')),
                ('textarea_rows', models.PositiveIntegerField(blank=True, null=True)),
            ],
            bases=('formulaic.Field',),
        ),
        migrations.CreateModel(
            name='SubmissionKeyValue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(db_index=True, max_length=200)),
                ('value_charfield', models.CharField(blank=True, max_length=500, null=True)),
                ('value_textfield', models.TextField(blank=True, null=True)),
                ('field', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='formulaic.Field')),
                ('submission', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='values', to='formulaic.Submission')),
            ],
        ),
        migrations.CreateModel(
            name='RuleResult',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('action', models.CharField(choices=[('show', 'Show'), ('hide', 'Hide'), ('require', 'Require (Override)'), ('optional', 'Optional (Override)'), ('change-option-group', 'Change Option Group')], max_length=50)),
                ('field', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='formulaic.Field')),
                ('option_group', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='formulaic.OptionGroup')),
                ('rule', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='results', to='formulaic.Rule')),
            ],
        ),
        migrations.CreateModel(
            name='RuleCondition',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('position', models.PositiveIntegerField()),
                ('operator', models.CharField(choices=[('is', 'is'), ('is_not', 'is_not'), ('contains', 'contains'), ('does_not_contain', 'does_not_contain'), ('begins_with', 'begins_with'), ('ends_with', 'ends_with'), ('greater_than', 'greater_than'), ('less_than', 'less_than'), ('any_selected', 'any_selected'), ('all_selected', 'all_selected')], max_length=30)),
                ('value_string', models.TextField()),
                ('field', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='formulaic.Field')),
                ('rule', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='conditions', to='formulaic.Rule')),
            ],
            options={
                'ordering': ('position',),
            },
        ),
        migrations.AddField(
            model_name='optiongroup',
            name='list',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='groups', to='formulaic.OptionList'),
        ),
        migrations.AddField(
            model_name='optiongroup',
            name='options',
            field=models.ManyToManyField(related_name='groups', to='formulaic.Option'),
        ),
        migrations.AddField(
            model_name='option',
            name='list',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='formulaic.OptionList'),
        ),
        migrations.AddField(
            model_name='form',
            name='privacy_policy',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='formulaic.PrivacyPolicy'),
        ),
        migrations.AddField(
            model_name='field',
            name='form',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='formulaic.Form'),
        ),
        migrations.CreateModel(
            name='DisplayCondition',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('values', models.TextField()),
                ('value_option', models.CharField(choices=[('is', 'is'), ('is_not', 'is not')], max_length=15)),
                ('affected_field', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='affecting_conditions', to='formulaic.Field')),
                ('watched_field', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='watching_conditions', to='formulaic.Field')),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='field',
            unique_together={('form', 'slug')},
        ),
        migrations.CreateModel(
            name='ChoiceField',
            fields=[
                ('field_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='formulaic.Field')),
                ('minimum_selections', models.PositiveIntegerField(blank=True, null=True)),
                ('maximum_selections', models.PositiveIntegerField(blank=True, null=True)),
                ('default_options_string', models.CharField(blank=True, max_length=200, null=True)),
                ('default_text', models.CharField(blank=True, max_length=200, null=True)),
                ('option_group', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='formulaic.OptionGroup')),
                ('option_list', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='formulaic.OptionList')),
            ],
            bases=('formulaic.Field',),
        ),
    ]
